/*
 *	A collection of file utility routines used by npasswd
 */

#include "npasswd.h"

#ifndef lint
static char sccsid[] = "@(#)file_util.c	1.1 03/26/98 (cc.utexas.edu) /usr/share/src/private/ut/share/bin/passwd/V2.0/src/Common/SCCS/s.file_util.c";
#endif

/*
 * File locking retry parameters
 */
#ifndef	FLOCK_TRIES
# define	FLOCK_TRIES 4
#endif
#ifndef	FLOCK_CYCLE
# define	FLOCK_CYCLE 2
#endif

Config_Value int	LockTries = FLOCK_TRIES;	/* How many times */
Config_Value int	LockCycle = FLOCK_CYCLE;	/* How long to wait */

/*
 * FileSizeDiff
 *	Find size difference between two files
 * Usage
 *	diff = FileSizeDiff(file1, file2);
 * Returns
 *	(size of <file2>) - (size of <file1>) bytes
 */
public size_t
FileSizeDiff(fn1, fn2)
	char	*fn1,	
		*fn2;
{
	struct stat	stb;	/* ACME Scratch Storage */
	size_t	s1;		/* ACME Scratch Storage */

	if (stat(fn1, &stb) < 0)
		return(-1);
	s1 = stb.st_size;
	if (stat(fn2, &stb) < 0)
		return(-1);
	return(s1 - stb.st_size);
}

/*
 * MakeLockTemp
 *	Create a file with exclusive use checking
 * Usage
 *	fd = MakeLockTemp(file);
 * Returns
 *	File descriptor of created file
 * Error
 *	Aborts if file cannot be created, usually because
 *	the file exists.
 */
public int
MakeLockTemp(name)
	char	*name;
{
	int	fd;	/* ACME Scratch Storage */
	int	cnt;	/* Retry counter */

	for (cnt = 0; cnt <= LockTries; cnt++) {
		if ((fd = open(name, O_WRONLY|O_CREAT|O_EXCL, 0600)) >= 0) 
			return(fd);
		debug(DB_DETAIL, "MakeLockTemp \"%s\" cycle %d\n", name, cnt);
		(void) sleep(LockCycle);
	}
	perror("Tempfile create");
	logdie("Cannot create temp file \"%s\"\n", name);
}

#if	!(defined(HAS_FCHMOD) && defined(HAS_FCHMOD))
/*
 * FixPwFileMode
 *	Set protection for passwd/shadow files
 * Usage
 *	FixPwFileMode(filename, "passwd" | "shadow");
 *
 * Used only when fchmod() and fchown() do not exist
 */
public void
FixPwFileMode(file, type)
	char	*file,
		*type;
{
	(void) chown(file, PASSWD_UID, PASSWD_GID);
	(void) chmod(file, PASSWD_MODE);
	if (strcmp(type, "shadow") == 0)
		(void) chmod(file, SHADOW_MODE);
}
#endif

/* End file_util.c */
